﻿/* 
 * Project:     Loader program for 9S12 microcontroller
 * Company:     Val Gretchev Engineering
 * Author:      Val Gretchev
 * Created:     January 2012
 * 
 * Notes:       Re-write of the program previously written in Delphi 7
 *
 *
 *
 *
 *
 *
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.IO.Ports;
using System.Threading;




namespace Load9S12TFR
{
    #region Variables

    // Enumerations
    public enum LogMsgType { Incoming, Outgoing, Normal, Warning, Error };
    
    public partial class Form1 : Form
    {
        // Constants
        private const char STX = (char)2;       // Start Of Text control character
        private const char ETX = (char)3;       // End Of Text control character
        private const char XON = (char)17;
        private const char XOFF = (char)19;
        private const int LengthFIFO = 1024;    // Length of receive FIFO
        private const int LengthLine = 256;     // Length of receive line
        
        // Various colors for logging info
        private Color[] LogMsgTypeColor = { Color.Blue, Color.Green, Color.Black, Color.Orange, Color.Red };

        // Global variables
        private char[] ReceiveFIFO = new char[LengthFIFO];      // Receive FIFO for incoming characters
        private int ReceiveInPtrFIFO;                           // Pointer for input characters
        private int ReceiveOutPtrFIFO;                          // Pointer for output characters
        private char[] ReceiveBfr = new char[LengthLine];       // Working buffer for a receive message
        private int ReceiveBfrPtr;                              // Pointer for receive buffer
        private bool ResponseExpected = false;                  // Flag to expect a response to a command

        // Array of bytes to be programmed into flash
        private byte[] Mem = new byte[45056];                   // Memory for arranging the flash data

        





        
        public Form1()
        {
            InitializeComponent();
        }

    #endregion


        #region Threads



        // This thread handles the TXD indicator light
        void HandleTXD()
        {
            /*
            while (true)
            {
                if ((serialPort1.IsOpen == true) && (serialPort1.BytesToWrite > 0))
                {
                    label3.BackColor = Color.Red;
                    Thread.Sleep(150);
                }

                label3.BackColor = Color.Black;
                Application.DoEvents();
                //Thread.Sleep(1);
            }
            */
            /*
            label3.BackColor = Color.Red;
            Thread.Sleep(150);
            label3.BackColor = Color.Black;
            */

            label3.BackColor = Color.Red;
            Thread.Sleep(150);
            while ((serialPort1.IsOpen == true) && (serialPort1.BytesToWrite > 0))
            {
                label3.BackColor = Color.Red;
                Thread.Sleep(50);
            }
            label3.BackColor = Color.Black;
            
        }

        // This thread handles the RXD indicator light
        void HandleRXD()
        {
            /*
            while (true)
            {
                if ((serialPort1.IsOpen == true) && (serialPort1.BytesToRead > 0))
                {
                    label7.BackColor = Color.Red;
                    Thread.Sleep(150);
                }

                label7.BackColor = Color.Black;
                Application.DoEvents();
                //Thread.Sleep(1);
            }
            */
            /*
            label7.BackColor = Color.Red;
            Thread.Sleep(150);
            label7.BackColor = Color.Black;
            */

            label7.BackColor = Color.Red;
            Thread.Sleep(150);
            while ((serialPort1.IsOpen == true) && (serialPort1.BytesToRead > 0))
            {
                label7.BackColor = Color.Red;
                Thread.Sleep(50);
            }
            label7.BackColor = Color.Black;
            
        }

        #endregion


        #region Common Methods

        private void setIndicatorLights()
        {
            if (serialPort1.DtrEnable == true)          
            {
                label1.BackColor = Color.Yellow;        // Data Terminal Ready indicator
            }
            else
            {
                label1.BackColor = Color.Black;
            }

            if (serialPort1.RtsEnable == true)          
            {
                label2.BackColor = Color.Yellow;        // Request To Send indicator
            }
            else
            {
                label2.BackColor = Color.Black;
            }

            if (serialPort1.DsrHolding == true)         
            {
                label4.BackColor = Color.Lime;          // Data Set Ready indicator
            }
            else
            {
                label4.BackColor = Color.Black;
            }

            if (serialPort1.CtsHolding == true)         
            {
                label5.BackColor = Color.Lime;          // Clear To Send indicator
            }
            else
            {
                label5.BackColor = Color.Black;
            }

            if (serialPort1.CDHolding == true)          
            {
                label6.BackColor = Color.Lime;          // Data Carrier Detect indicator
            }
            else
            {
                label6.BackColor = Color.Black;
            }
        }

        private char GetCharInBuf()
        {
            char CharBuf;

            CharBuf = ReceiveFIFO[ReceiveOutPtrFIFO++];
            if (ReceiveOutPtrFIFO >= LengthFIFO)
            {
                ReceiveOutPtrFIFO = 0;
            }
            return CharBuf;
        }

        private int WaitForResponse(int DelayTime)
        {
            bool Done = false;
            int Result = -1;
            timer1.Enabled = false;
            timer1.Interval = DelayTime;
            timer1.Enabled = true;
            //ReceiveBfr[0] = '\0';
            //Array.Clear(ReceiveBfr, 0, ReceiveBfr.Length);

            while ((!Done) && (timer1.Enabled == true))
            {
                while (ReceiveOutPtrFIFO == ReceiveInPtrFIFO)
                {
                    Application.DoEvents();
                    if (timer1.Enabled == false)
                    {
                        break;
                    }
                }

                if (ReceiveOutPtrFIFO != ReceiveInPtrFIFO)
                {
                    char RxChar = GetCharInBuf();

                    switch (RxChar)
                    {
                        case '\n':
                            ReceiveBfr[ReceiveBfrPtr++] = RxChar;
                            //ReceiveLine[ReceiveCharPtr++] = Convert.ToChar(0);
                            ReceiveBfr[ReceiveBfrPtr++] = '\0';
                            ReceiveBfrPtr = 0;
                            Result = 0;
                            Done = true;
                            ResponseExpected = false;
                            timer1.Enabled = false;
                            string line = new string(ReceiveBfr);
                            break;

                        case '\r':
                            break;

                        default:
                            if (ReceiveBfrPtr < (LengthLine - 5))
                                ReceiveBfr[ReceiveBfrPtr++] = RxChar;
                            break;
                    }
                }
            }
            
            switch (Result)
            {
                case -1:
                    WriteStatusMsg(Color.Red, "Timeout on port");
                    //return;
                    break;
                case 0:
                    WriteStatusMsg(Color.Black, "Received response");
                    break;
            }
            button1.Enabled = true;
            return Result;
        }

        private void SendCommand(string CmdStr)
        {
            ResponseExpected = true;
            serialPort1.Write(CmdStr);
            Thread handleTxd = new Thread(new ThreadStart(HandleTXD));
            handleTxd.IsBackground = true;
            handleTxd.Start();
            
            //string SendData = CmdStr.Replace("\n", "");
            //toolStripStatusLabel1.ForeColor = Color.Black;
            //toolStripStatusLabel1.Text = "";
            //toolStripStatusLabel1.Text = "Sending: " + SendData;
        }

        // Writes status messages
        private void WriteStatusMsg(Color SetColor, string Msg)
        {
            toolStripStatusLabel1.ForeColor = SetColor;
            toolStripStatusLabel1.Text = Msg;
        }


        /* Author:     Noah Coad, http://coad.net/noah */
        private void Log(LogMsgType msgtype, string msg)
        {
            richTextBox2.Invoke(new EventHandler(delegate
            {
                richTextBox2.SelectedText = string.Empty;
                richTextBox2.SelectionFont = new Font(richTextBox2.SelectionFont, FontStyle.Bold);
                richTextBox2.SelectionColor = LogMsgTypeColor[(int)msgtype];
                richTextBox2.AppendText(msg);
                richTextBox2.ScrollToCaret();
            }));
        }

        /* Author:     Noah Coad, http://coad.net/noah */
        private string[] OrderedPortNames()
        {
            // Just a placeholder for a successful parsing of a string to an integer
            int num;

            // Order the serial port names in numberic order (if possible)
            return SerialPort.GetPortNames().OrderBy(a => a.Length > 3 && int.TryParse(a.Substring(3), out num) ? num : 0).ToArray();
        }

        /* Author:     Noah Coad, http://coad.net/noah */
        /// <summary> Convert a string of hex digits (ex: E4CAB2) to a byte array. </summary>
        /// <param name="s"> The string containing the hex digits. </param>
        /// <returns> Returns an array of bytes. </returns>
        private byte[] HexStringToByteArray(string s)
        {
            if (s.Length % 2 != 0)
            {
                MessageBox.Show(this, "String must have\neven number of bytes.", "Error in format", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return null;
            }
            byte[] buffer = new byte[s.Length / 2];
            for (int i = 0; i < s.Length; i += 2)
                buffer[i / 2] = (byte)Convert.ToByte(s.Substring(i, 2), 16);
            return buffer;
        }

        /* Author:     Noah Coad, http://coad.net/noah */
        /// <summary> Converts an array of bytes into a formatted string of hex digits (ex: E4CAB2)</summary>
        /// <param name="data"> The array of bytes to be translated into a string of hex digits. </param>
        /// <returns> Returns a string of hex digits without spacing. </returns>
        private string ByteArrayToHexString(byte[] data)
        {
            StringBuilder sb = new StringBuilder(data.Length * 2);
            foreach (byte b in data)
                sb.Append(Convert.ToString(b, 16).PadLeft(2, '0'));
            return sb.ToString().ToUpper();
        }
        
        #endregion


        #region Application
        
        private void ProgramFlash()
        {
            string OutBuf;
            int Status;
            int Address;
            bool DataPresent;
            byte[] Work = new byte[64];
            
            for (int i = 0; i < Mem.Length; i++)    // Initialize memory array to erased condition of flash
            {
                Mem[i] = 0xFF;
            }

            int NumLines = richTextBox1.Lines.Length;   // Get the line length of the file
            //textBox8.Text = NumLines.ToString();        // Diagnostic

            if (NumLines > 0)   // Eliminate an empty richTextBox
            {
                foreach (string line in richTextBox1.Text.Split('\n'))
                {
                    if (line.Length > 0)    // Eliminate any empty lines
                    {
                        string recordType = line.Substring(0, 2);   // Extract the S0, S1, S9 type
                        if (recordType == "S1")                     // Only interested in S1 type
                        {
                            // Parse the line into separate strings
                            string recordLength = line.Substring(2, 2);         // Extract the number of bytes in the line
                            string addressField = line.Substring(4, 4);         // Extract the address field
                            string dataField = line.Substring(8, line.Length - 10); // Extract the data characters
                            string checksum = line.Substring(line.Length - 2, 2);   // Extract the checksum
                            //richTextBox2.AppendText(recordLength + " " + addressField + " " + dataField + " " + checksum + "\n");   // Diagnostic                                                        

                            // Convert strings to numbers
                            int recordSize = Convert.ToInt32(recordLength, 16); // Convert record length to a number
                            int dataCount = recordSize - 3;                     // Subtract length of address and checksum
                            int nibbleCount = dataCount * 2;                    // Nibble count is bytes times 2
                            int address = Convert.ToInt32(addressField, 16);    // Convert address field to a number
                            address = address - 0x4000;                         // Subtract flash address offset

                            // Convert data field to bytes
                            byte[] buffer = new byte[dataField.Length / 2];
                            buffer = HexStringToByteArray(dataField);
                            //richTextBox2.AppendText(ByteArrayToHexString(buffer) + "\n");   // Diagnostic

                            // Load memory array
                            for (int i = 0; i < dataCount; i++)
                            {
                                Mem[address + i] = buffer[i];
                                //richTextBox2.AppendText((address + i).ToString() + " " + i.ToString() + "\n");    // Diagnostic
                            }
                            // At this point the memory array has been loaded with file data
                        }
                    }
                }

                toolStripProgressBar1.Value = 0;
                richTextBox2.Clear();
                WriteStatusMsg(Color.Blue, "Erasing flash at 0x4000, please wait");

                OutBuf = STX + "07" + ETX;
                Log(LogMsgType.Outgoing, OutBuf + "\n");
                SendCommand(OutBuf);
                Status = WaitForResponse(5000);
                if (Status != 0)
                    return;

                WriteStatusMsg(Color.Blue, "Erasing flash at 0xC000, please wait");

                OutBuf = STX + "08" + ETX;
                Log(LogMsgType.Outgoing, OutBuf + "\n");
                SendCommand(OutBuf);
                Status = WaitForResponse(5000);
                if (Status != 0)
                    return;

                WriteStatusMsg(Color.Blue, "Writing flash in address 0x4000 to 0x7FFF, please wait");
                Address = 0;
                while (Address < (16384 - 1))
                {
                    DataPresent = false;
                    OutBuf = "";
                    for (int i = 0; i < 64; i++)
                    {
                        Work[i] = Mem[Address + i];
                        if (Work[i] != 0xFF)
                        {
                            DataPresent = true;
                        }
                    }
                    if (DataPresent)
                    {
                        OutBuf = ByteArrayToHexString(Work);
                        OutBuf = STX + "01" + Convert.ToString((Address + 0x4000), 16).ToUpper() + Convert.ToString(64, 16).ToUpper() + OutBuf + ETX;
                        Log(LogMsgType.Outgoing, OutBuf + "\n");
                        SendCommand(OutBuf);
                        Status = WaitForResponse(1000);
                        if (Status != 0)
                            return;
                    }
                    Address += 64;
                    toolStripProgressBar1.Value = (Address * 100) / 45056;
                }
                toolStripProgressBar1.Value = 100;

                WriteStatusMsg(Color.Blue, "Writing flash in address 0xC000 to 0xEFFF, please wait");
                Address = 32768;
                while (Address < (45056 - 1))
                {
                    DataPresent = false;
                    OutBuf = "";
                    for (int i = 0; i < 64; i++)
                    {
                        Work[i] = Mem[Address + i];
                        if (Work[i] != 0xFF)
                        {
                            DataPresent = true;
                        }
                    }
                    if (DataPresent)
                    {
                        OutBuf = ByteArrayToHexString(Work);
                        OutBuf = STX + "01" + Convert.ToString((Address + 0x4000), 16).ToUpper() + Convert.ToString(64, 16).ToUpper() + OutBuf + ETX;
                        Log(LogMsgType.Outgoing, OutBuf + "\n");
                        SendCommand(OutBuf);
                        Status = WaitForResponse(1000);
                        if (Status != 0)
                            return;
                    }
                    Address += 64;
                    toolStripProgressBar1.Value = (Address * 100) / 45056;
                }
                toolStripProgressBar1.Value = 100;
            }
            else
            {
                MessageBox.Show(this, "Please load a file", "Missing File Data", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
         }

        private void ReadMem()
        {
            int Address;
            string OutputBuf;
            byte[] Work = new byte[64];
            
            richTextBox2.Clear();
            Address = 0;
            while (Address < (45056 - 1))
            {
                OutputBuf = "";
                for (int i = 0; i < 64; i++)
                {
                    Work[i] = Mem[Address + i];
                }

                OutputBuf = ByteArrayToHexString(Work);
                OutputBuf = Convert.ToString((Address + 0x4000), 16).ToUpper() + " " + OutputBuf;
                Log(LogMsgType.Normal, OutputBuf + "\n");
                Address += 64;
            }
        }

        private void GetStatus()
        {
            string OutBuf;
            int Status;

            WriteStatusMsg(Color.Black, "");
            OutBuf = STX + "04" + ETX;
            Log(LogMsgType.Outgoing, OutBuf + "\n");
            SendCommand(OutBuf);
            Status = WaitForResponse(1000);
        }

        private int ReadFash(int StartAddr, int LineCount)
        {
            string OutBuf;
            int Status = 0;

            OutBuf = STX + "02" + Convert.ToString(StartAddr, 16).ToUpper().PadLeft(4, '0') + Convert.ToString(LineCount, 16).ToUpper().PadLeft(2, '0') + ETX;
            Log(LogMsgType.Outgoing, OutBuf + "\n");
            WriteStatusMsg(Color.Black, "Retreiving flash code");
            SendCommand(OutBuf);
            toolStripProgressBar1.Value = 0;
            for (int i = 0; i < LineCount; i++)
            {
                ResponseExpected = true;
                Status = WaitForResponse(1000);
                if (Status != 0)
                {
                    WriteStatusMsg(Color.Black, "Status not 0");
                    //break;
                }
                toolStripProgressBar1.Value = ((i * 100) / LineCount);
            }
            
            toolStripProgressBar1.Value = 100;
            return Status;
        }

        private void Execute()
        {
            string OutBuf;
            int Status;

            WriteStatusMsg(Color.Black, "");
            OutBuf = STX + "05" + ETX;
            Log(LogMsgType.Outgoing, OutBuf + "\n");
            SendCommand(OutBuf);
            Status = WaitForResponse(1000);
        }

        #endregion


        #region Events

        private void Form1_Load(object sender, EventArgs e)
        {
            string[] ports = OrderedPortNames();    // This will order the port names


            // Enter each port name into the combo box.
            foreach (string port in ports)
            {
                comboBox1.Items.Add(port);
            }

            if (comboBox1.Items.Count > 0)
            {
                comboBox1.SelectedIndex = 0;    // Automatically select the first com port in the list
                serialPort1.Close();
                serialPort1.PortName = comboBox1.Text;
                //serialPort1.BaudRate = 19200;
                serialPort1.BaudRate = 9600;
                serialPort1.DtrEnable = true;
                serialPort1.RtsEnable = true;
                serialPort1.Open();
                setIndicatorLights();
                WriteStatusMsg(Color.Black, "Ready");
                ReceiveInPtrFIFO = 0;
                ReceiveOutPtrFIFO = 0;

                // Define and start the TXD thread handler
                //Thread handleTxd = new Thread(new ThreadStart(HandleTXD));
                //handleTxd.IsBackground = true;
                //handleTxd.Start();

                // Define and start the RXD thread handler
                //Thread handleRxd = new Thread(new ThreadStart(HandleRXD));
                //handleRxd.IsBackground = true;
                //handleRxd.Start();
            }
            else
            {
                //MessageBox.Show("No ports available");
                WriteStatusMsg(Color.Red, "No port available");
            }
        }

        private void serialPort1_PinChanged(object sender, SerialPinChangedEventArgs e)
        {
            setIndicatorLights();
        }

        private void comboBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            serialPort1.Close();
            serialPort1.PortName = comboBox1.Text;
            serialPort1.Open();
        }

        private void label1_Click(object sender, EventArgs e)
        {
            if (serialPort1.DtrEnable == true)
            {
                serialPort1.DtrEnable = false;
            }
            else
            {
                serialPort1.DtrEnable = true;
            }
            setIndicatorLights();
        }

        private void label2_Click(object sender, EventArgs e)
        {
            if (serialPort1.RtsEnable == true)
            {
                serialPort1.RtsEnable = false;
            }
            else
            {
                serialPort1.RtsEnable = true;
            }
            setIndicatorLights();
        }

        private void serialPort1_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            string data = serialPort1.ReadExisting();

            // This prevents unsolicited messages from entering into the receive FIFO
            if (ResponseExpected)
            {
                for (int i = 0; i < data.Length; i++)
                {
                    ReceiveFIFO[ReceiveInPtrFIFO++] = data[i];
                    if (ReceiveInPtrFIFO >= LengthFIFO)
                    {
                        ReceiveInPtrFIFO = 0;
                    }
                }
            }

            // Log all messages to the terminal display
            data = data.Replace("\r", "");  // This prevents the rich text box from misfiring
            //data = data.Replace("\n", "");
            Log(LogMsgType.Incoming, data);

            Thread handleRxd = new Thread(new ThreadStart(HandleRXD));
            handleRxd.IsBackground = true;
            handleRxd.Start();
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            timer1.Enabled = false;
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            serialPort1.Close();    // Probably not required
        }

        private void exitToolStripMenuItem_Click_1(object sender, EventArgs e)
        {
            Close();
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // Determine whether the user selected a file from the OpenFileDialog.
            if (openFileDialog1.ShowDialog() == System.Windows.Forms.DialogResult.OK &&
                openFileDialog1.FileName.Length > 0)
            {
                // Clear the rich text box
                richTextBox1.Clear();

                // Progress bar to starting value
                toolStripProgressBar1.Value = 0;

                // Load the contents of the file into the RichTextBox.
                richTextBox1.LoadFile(openFileDialog1.FileName, RichTextBoxStreamType.PlainText);
            }
        }

        private void button1_Click(object sender, EventArgs e)
        {
            if (textBox1.Text == "")
            {
                WriteStatusMsg(Color.Red, "Command is blank");
            }
            else
            {
                button1.Enabled = false;
                WriteStatusMsg(Color.Black, "");
                Log(LogMsgType.Outgoing, textBox1.Text + "\n");
                //SendCommand(textBox1.Text + "\r");
                SendCommand(textBox1.Text + "\n");
                int Status = WaitForResponse(2000);
            }
        }

        #endregion

        

        private void programToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ProgramFlash();
        }

        private void readMemToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ReadMem();
        }

        private void getStatusToolStripMenuItem_Click(object sender, EventArgs e)
        {
            GetStatus();
        }

        private void readFlashToolStripMenuItem_Click(object sender, EventArgs e)
        {
            richTextBox2.Clear();
            
            /*
            if (ReadFash(16384, 128) != 0)
                return;
             if (ReadFash(24576, 128) != 0)
                return;
            if (ReadFash(49152, 128) != 0)
                return;
            if (ReadFash(57344, 64) != 0)
                return;
            */
            ReadFash(16384, 128);
            ReadFash(24576, 128);
            ReadFash(49152, 128);
            ReadFash(57344, 64);
        }

        private void readLoaderToolStripMenuItem_Click(object sender, EventArgs e)
        {
            richTextBox2.Clear();
            ReadFash(61440, 64);
        }

        private void executeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Execute();
        }

        private void clearPanel1ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            richTextBox1.Clear();
        }

        private void clearPanel2ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            richTextBox2.Clear();
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            using (AboutBox1 box = new AboutBox1())
                box.ShowDialog();
        }

        private void button2_Click(object sender, EventArgs e)
        {
            byte[] c = new byte[] { 0x11 };
 
            //Log(LogMsgType.Outgoing, "Sending XON\n");
            WriteStatusMsg(Color.Blue, "Sending XON");
            //serialPort1.Write(XON + "\0");
            serialPort1.Write(c, 0, 1);

            Thread handleTxd = new Thread(new ThreadStart(HandleTXD));
            handleTxd.IsBackground = true;
            handleTxd.Start();
        }

        private void goToLoaderToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            Log(LogMsgType.Outgoing, "Sending COP reset command\n");
            //serialPort1.Write("z\n");
            SendCommand("z\n");
        }

        private void button3_Click(object sender, EventArgs e)
        {
            byte[] c = new byte[] { 0x13 };
            
            //Log(LogMsgType.Outgoing, "Sending XOFF\n");
            WriteStatusMsg(Color.Red, "Sending XOFF");
            //serialPort1.Write(XOFF + "\0");
            serialPort1.Write(c, 0, 1);

            Thread handleTxd = new Thread(new ThreadStart(HandleTXD));
            handleTxd.IsBackground = true;
            handleTxd.Start();
        }



    }   
}
